@@ -4,7 +4,13 @@ Module for jenkinsapi Job
 
 import json
 import logging
-import urlparse
+
+try:
+    import urlparse
+except ImportError:
+    # Python3
+    import urllib.parse as urlparse
+
 import xml.etree.ElementTree as ET
 from collections import defaultdict
 from time import sleep
@@ -99,7 +105,7 @@ class Job(JenkinsBase, MutableJenkinsThing):
         Jenkins API loads the first 100 builds and thus may not contain all builds
         information. This method checks if all builds are loaded in the data object
         and updates it with the missing builds if needed.'''
-        if not (data.get("builds") and data.get("firstBuild")):
+        if not data.get("builds"):
             return data
         # do not call _buildid_for_type here: it would poll and do an infinite loop
         oldest_loaded_build_number = data["builds"][-1]["number"]
@@ -132,20 +138,29 @@ class Job(JenkinsBase, MutableJenkinsThing):
         return "%s/buildWithParameters" % self.baseurl
 
     @staticmethod
-    def _mk_json_from_build_parameters(build_params):
+    def _mk_json_from_build_parameters(build_params, file_params=None):
         """
-        Build parameters must be submitted in a particular format - Key-Value pairs would be
-        far too simple, no no! Watch and read on and behold!
+        Build parameters must be submitted in a particular format
+        Key-Value pairs would be far too simple, no no!
+        Watch and read on and behold!
         """
         assert isinstance(
             build_params, dict), 'Build parameters must be a dict'
-        return {'parameter': [
-            {'name': k, 'value': v} for k, v in build_params.iteritems()
-        ]}
+
+        build_p = [{'name': k, 'value': v}
+                   for k, v in build_params.items()]
+        out = {'parameter': build_p}
+        if file_params:
+            file_p = [{'name': k, 'file': k}
+                      for k in file_params.keys()]
+            out['parameter'].extend(file_p)
+
+        return out
 
     @staticmethod
-    def mk_json_from_build_parameters(build_params):
-        to_json_structure = Job._mk_json_from_build_parameters(build_params)
+    def mk_json_from_build_parameters(build_params, file_params=None):
+        to_json_structure = Job._mk_json_from_build_parameters(build_params,
+                                                               file_params)
         return json.dumps(to_json_structure)
 
     def invoke(self, securitytoken=None, block=False, skip_if_running=False, invoke_pre_check_delay=3,
@@ -183,6 +198,11 @@ class Job(JenkinsBase, MutableJenkinsThing):
                 self.get_jenkins_obj()))
 
             url = self.get_build_triggerurl()
+            # If job has file parameters - it must be triggered
+            # using "/build", not by "/buildWithParameters"
+            # "/buildWithParameters" will ignore non-file parameters
+            if files:
+                url = "%s/build" % self.baseurl
 
             if cause:
                 build_params['cause'] = cause
@@ -190,6 +210,7 @@ class Job(JenkinsBase, MutableJenkinsThing):
             if securitytoken:
                 params['token'] = securitytoken
 
+            build_params['json'] = self.mk_json_from_build_parameters(build_params, files)
             data = build_params
 
             response = self.jenkins.requester.post_and_confirm_status(
@@ -368,6 +389,16 @@ class Job(JenkinsBase, MutableJenkinsThing):
         url = self.get_build_dict()[buildnumber]
         return Build(url, buildnumber, job=self)
 
+    def get_build_metadata(self, buildnumber):
+        """
+        Get the build metadata for a given build number. For large builds with
+        tons of tests, this method is faster than get_build by returning less
+        data.
+        """
+        assert type(buildnumber) == int
+        url = self.get_build_dict()[buildnumber]
+        return Build(url, buildnumber, job=self, depth=0)
+
     def __getitem__(self, buildnumber):
         return self.get_build(buildnumber)
 
@@ -494,8 +525,13 @@ class Job(JenkinsBase, MutableJenkinsThing):
         Also refresh the ElementTree object since the config has changed
         """
         url = self.get_config_xml_url()
-        if isinstance(config, unicode):
-            config = str(config)
+        try:
+            if isinstance(config, unicode):  # pylint: disable=undefined-variable
+                config = str(config)
+        except NameError:
+            # Python3 already a str
+            pass
+
         response = self.jenkins.requester.post_url(url, params={}, data=config)
         self._element_tree = ET.fromstring(config)
         return response.text
@@ -509,7 +545,7 @@ class Job(JenkinsBase, MutableJenkinsThing):
         try:
             for j in self._data['downstreamProjects']:
                 downstream_jobs.append(
-                    self.get_jenkins_obj().get_job(j['name']))
+                    self.get_jenkins_obj()[j['name']])
         except KeyError:
             return []
         return downstream_jobs
@@ -606,7 +642,10 @@ class Job(JenkinsBase, MutableJenkinsThing):
         return [param['name'] for param in self.get_params()]
 
     def has_params(self):
-        return len(self._data['actions']) > 0
+        """
+        If job has parameters, returns True, else False
+        """
+        return any("parameterDefinitions" in a for a in self._data["actions"])
 
     def has_queued_build(self, build_params):
         """Returns True if a build with build_params is currently queued."""
